import logging

import django.conf
from django.contrib.auth import load_backend
from allauth.account.models import EmailAddress
from django_auth_ldap.config import LDAPSearch
from django_mailman3.lib.mailman import get_mailman_user, sync_email_addresses, update_preferred_address

logger = logging.getLogger(__name__)

ldap_backend = load_backend('django_auth_ldap.backend.LDAPBackend')

def get_ldap_connection():
	conn = ldap_backend.ldap.initialize(ldap_backend.settings.SERVER_URI, bytes_mode=False)
	for opt, value in ldap_backend.settings.CONNECTION_OPTIONS.items():
		conn.set_option(opt, value)
	if ldap_backend.settings.START_TLS:
		conn.start_tls_s()
	conn.simple_bind_s(ldap_backend.settings.BIND_DN, ldap_backend.settings.BIND_PASSWORD)
	return conn

def execute_ldap_search_without_hiding_errors(search, connection, filterargs=(), escape=True):
	'''If an LDAPError occurs during search.execute, it logs the error and returns
	an empty list of results. This behavor is indistinguishable from a query
	with no results.'''
	if not isinstance(search, LDAPSearch):
		raise NotImplementedError(f'{type(search)} is not supported by postorius_ldap_membership_management')
	# This is a copy of django_auth_ldap.config.LDAPSearch.execute without the
	# ldap.LDAPError eating try-except block
	if escape:
		filterargs = search._escape_filterargs(filterargs)
	filterstr = search.filterstr % filterargs
	results = connection.search_s(search.base_dn, search.scope, filterstr, search.attrlist)
	return search._process_results(results)

def populate_user(username):
	'''Create or update user from LDAP, including it's django-allauth email
	addresses and Mailman backend user'''
	# Caution: If a user has no verified primary email address (an EmailAddress
	# object with verified=True and primary=True), django-allauth allows the user
	# to set his primary email address and also User.email to any unverified
	# address. This is a security problem, because Postorius and Hyperkitty use
	# User.email to lookup the corresponding Mailman backend user and thus for
	# permission checking. This causes a race-condition in the following code
	# that we can't do anything about. ldap_backend.populate_user creates (and
	# saves) a new user. We create the EmailAddress objects afterwards. If this
	# fails or does not happen for some reason, we end up with a user without a
	# primary email address.
	user = ldap_backend.populate_user(username)
	if user is None:
		# ldap_backend.populate_user hides LDAP errors and returns None if
		# they occur. It also returns None if the user is not found in the LDAP
		# directory.
		return None
	if 'last_name' not in django.conf.settings.AUTH_LDAP_USER_ATTR_MAP:
		user.last_name = ''

	primary_changed = False

	user_email_objs = EmailAddress.objects.filter(user=user)
	primary_email_obj = ([obj for obj in user_email_objs if obj.primary] or [None])[0]
	ldap_email = user.ldap_user.attrs[django.conf.settings.AUTH_LDAP_USER_ATTR_MAP['email']][0]
	ldap_email_obj = ([obj for obj in user_email_objs if obj.email == ldap_email] or [None])[0]
	if ldap_email_obj is None:
		ldap_email_obj = EmailAddress(user=user, email=ldap_email, verified=False, primary=False)
	ldap_email_obj.verified = True
	if primary_email_obj is None:
		ldap_email_obj.primary = True
		primary_email_obj = ldap_email_obj
		primary_changed = True
	ldap_email_obj.save()

	# ldap_backend.populate_user overwrites user.email with the address from LDAP
	user.email = primary_email_obj.email
	user.save()

	sync_email_addresses(user)
	if primary_changed:
		update_preferred_address(user, primary_email_obj)
	mm_user = get_mailman_user(user)
	if mm_user.display_name != user.get_full_name():
		mm_user.display_name = user.get_full_name()
		mm_user.save()
	return user
